"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

# import logging
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import COMMANDS, getConstName


class CommandPacket(Packet):
    '''
    Implementation based on description:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    def setup(self, length=0, count_header=1, body=''):
        if length == 0:
            return

        # self.salt=self.session.salt

        self._length = length
        self._seq = count_header
        self._body = body

        self.type = Datum('int<1>')
        buffer = body
        buffer = self.type.setFromBuff(buffer)

        if self.type.value == COMMANDS.COM_QUERY:
            self.sql = Datum('str<EOF>')
            self.sql.setFromBuff(buffer)
        else:
            self.data = Datum('str<EOF>')
            self.data.setFromBuff(buffer)

    def __str__(self):
        return str({
            'header': {'length': self.length, 'seq': self.seq},
            'type': getConstName(COMMANDS, self.type.value),
            'vars': self.__dict__
        })



import requests

import mysql.connector

from mindsdb.libs.constants.mindsdb import DATA_TYPES, DATA_SUBTYPES


class Mariadb():
    def __init__(self, config):
        self.config = config
        self.host = config['integrations']['mariadb']['host']
        self.port = config['integrations']['mariadb']['port']
        self.user = config['integrations']['mariadb']['user']
        self.password = config['integrations']['mariadb']['password']

        self.setup_mariadb()


    def _to_mariadb_table(self, stats):
        subtype_map = {
            DATA_SUBTYPES.INT: 'int',
            DATA_SUBTYPES.FLOAT: 'double',
            DATA_SUBTYPES.BINARY: 'bool',
            DATA_SUBTYPES.DATE: 'Date',
            DATA_SUBTYPES.TIMESTAMP: 'Datetime',
            DATA_SUBTYPES.SINGLE: 'VARCHAR(500)',
            DATA_SUBTYPES.MULTIPLE: 'VARCHAR(500)',
            DATA_SUBTYPES.IMAGE: 'VARCHAR(500)',
            DATA_SUBTYPES.VIDEO: 'VARCHAR(500)',
            DATA_SUBTYPES.AUDIO: 'VARCHAR(500)',
            DATA_SUBTYPES.TEXT: 'VARCHAR(500)',
            DATA_SUBTYPES.ARRAY: 'VARCHAR(500)'
        }

        column_declaration = []
        for name, column in stats.items():
            try:
                col_subtype = stats[name]['typing']['data_subtype']
                new_type = subtype_map[col_subtype]
                column_declaration.append(f' `{name}` {new_type} ')
            except Exception as e:
                print(e)
                print(f'Error: cant convert type {col_subtype} of column {name} to mariadb tpye')

        return column_declaration

    def _query(self, query):
        con = mysql.connector.connect(host=self.host, port=self.port, user=self.user, password=self.password)

        cur = con.cursor(dictionary=True,buffered=True)
        cur.execute(query)
        res = True
        try:
            res = cur.fetchall()
        except:
            pass
        con.commit()
        con.close()

        return res

    def _get_connect_string(self, table):
        user = self.config['api']['mysql']['user']
        password = self.config['api']['mysql']['password']
        host = self.config['api']['mysql']['host']
        port = self.config['api']['mysql']['port']

        if password is None or password == '':
            connect = f'mysql://{user}@{host}:{port}/mindsdb/{table}'
        else:
            connect = f'mysql://{user}:{password}@{host}:{port}/mindsdb/{table}'

        return connect

    def setup_mariadb(self):
        self._query('CREATE DATABASE IF NOT EXISTS mindsdb')

        connect = self._get_connect_string('predictors_mariadb')

        q = f"""
                CREATE TABLE IF NOT EXISTS mindsdb.predictors
                (name VARCHAR(500),
                status VARCHAR(500),
                accuracy VARCHAR(500),
                predict_cols VARCHAR(500),
                select_data_query VARCHAR(500),
                training_options VARCHAR(500)
                ) ENGINE=CONNECT TABLE_TYPE=MYSQL CONNECTION='{connect}';
        """
        print(f'Executing table creation query to create predictors list:\n{q}\n')
        self._query(q)

        q = f"""
            CREATE TABLE IF NOT EXISTS mindsdb.commands (
                command VARCHAR(500)
            ) ENGINE=CONNECT TABLE_TYPE=MYSQL CONNECTION='{connect}';
        """
        print(f'Executing table creation query to create command table:\n{q}\n')
        self._query(q)

    def register_predictor(self, name, stats):
        columns_sql = ','.join(self._to_mariadb_table(stats))
        columns_sql += ',`$select_data_query` varchar(500)'

        connect = self._get_connect_string(f'{name}_mariadb')

        q = f"""
                CREATE TABLE mindsdb.{name}
                ({columns_sql}
                ) ENGINE=CONNECT TABLE_TYPE=MYSQL CONNECTION='{connect}';
        """
        print(f'Executing table creation query to sync predictor:\n{q}\n')
        self._query(q)

    def unregister_predictor(self, name):
        q = f"""
            drop table if exists mindsdb.{name};
        """
        print(f'Executing table creation query to sync predictor:\n{q}\n')
        self._query(q)



from flask import request
from flask_restx import Resource, abort

from mindsdb.api.http.namespaces.configs.util import ns_conf


@ns_conf.route('/ping')
class Ping(Resource):
    @ns_conf.doc('get_ping')
    def get(self):
        '''Checks server avaliable'''
        return {'status': 'ok'}


@ns_conf.route('/shutdown')
class Shutdown(Resource):
    @ns_conf.doc('get_shutdown')
    def get(self):
        '''Shutdown server'''
        if request.host.startswith('127.0.0.1') or request.host.startswith('localhost'):
            func = request.environ.get('werkzeug.server.shutdown')
            if func is None:
                return '', 500
            func()
            return '', 200
        abort(403, "")



import json
import os
import mindsdb
import logging
import sys
import random

from mindsdb.api.http.namespaces.predictor import ns_conf as predictor_ns
from mindsdb.api.http.namespaces.datasource import ns_conf as datasource_ns
from mindsdb.api.http.namespaces.util import ns_conf as utils_ns
from mindsdb.api.http.initialize import initialize_flask, initialize_interfaces
from mindsdb.utilities.config import Config

def start(config):
    config = Config(config)
    port=47334
    host='0.0.0.0'
    debug=False

    if not logging.root.handlers:
        rootLogger = logging.getLogger()

        outStream = logging.StreamHandler(sys.stdout)
        outStream.addFilter(lambda record: record.levelno <= logging.INFO)
        rootLogger.addHandler(outStream)

        errStream = logging.StreamHandler(sys.stderr)
        errStream.addFilter(lambda record: record.levelno > logging.INFO)
        rootLogger.addHandler(errStream)


    mindsdb.CONFIG.MINDSDB_DATASOURCES_PATH = os.path.join(mindsdb.CONFIG.MINDSDB_STORAGE_PATH,'datasources')
    mindsdb.CONFIG.MINDSDB_TEMP_PATH = os.path.join(mindsdb.CONFIG.MINDSDB_STORAGE_PATH,'tmp')

    os.makedirs(mindsdb.CONFIG.MINDSDB_STORAGE_PATH, exist_ok=True)
    os.makedirs(mindsdb.CONFIG.MINDSDB_DATASOURCES_PATH, exist_ok=True)
    os.makedirs(mindsdb.CONFIG.MINDSDB_TEMP_PATH, exist_ok=True)
    #'''
    app, api = initialize_flask(config)
    initialize_interfaces(config, app)

    api.add_namespace(predictor_ns)
    api.add_namespace(datasource_ns)
    api.add_namespace(utils_ns)

    app.run(debug=debug, port=config['api']['http']['port'], host=config['api']['http']['host'])

if __name__ == '__main__':
    start()



import os
import json


def _in(ask, default, use_default):
    if use_default:
        return default

    user_input = input(f'{ask} (Default: {default})')
    if user_input is None or user_input == '':
        user_input = default

    if type(default) == int:
        user_input = int(user_input)
        
    return user_input

def auto_config(python_path,pip_path,predictor_dir,datasource_dir):
    config = {
        "debug": False
        ,"config_version": 1
        ,"python_interpreter": python_path
        ,"pip_path": pip_path
        ,"api": {
        }
        ,"integrations": {
          "clickhouse": {
              "enabled": False
          }
          ,"mariadb": {
              "enabled": False
          }
        }
        ,"interface":{
          "mindsdb_native": {
              "enabled": True
              ,"storage_dir": predictor_dir
          }
          ,"lightwood": {
               "enabled": True
          }
          ,"datastore": {
               "enabled": True
               ,"storage_dir": datasource_dir
          }
          ,"dataskillet": {
               "enabled": False
          }
        }
    }

    return config

def cli_config(python_path,pip_path,predictor_dir,datasource_dir,config_dir,use_default=False):
    config = auto_config(python_path,pip_path,predictor_dir,datasource_dir)

    http = _in('Enable HTTP API ? [Y/N]','Y',use_default)
    if http in ['Y','y']:
        config['api']['http'] = {}
        config['api']['http']['host'] = _in('HTTP interface host: ','0.0.0.0',use_default)
        config['api']['http']['port'] = _in('HTTP interface port: ','47334',use_default)

    mysql = _in('Enable MYSQL API ? [Y/N]','Y',use_default)
    if mysql in ['Y','y']:
        config['api']['mysql'] = {
            "certificate_path": "cert.pem"
            ,"log": {
                "format": "%(asctime)s - %(levelname)s - %(message)s",
                "folder": "logs/",
                "file": "mysql.log",
                "file_level": "INFO",
                "console_level": "INFO"
            }
            ,"datasources": []
        }
        config['api']['mysql']['host'] = _in('MYSQL interface host','127.0.0.1',use_default)
        config['api']['mysql']['port'] = _in('MYSQL interface port','47335',use_default)
        config['api']['mysql']['user'] = _in('MYSQL interface user','mindsdb',use_default)
        config['api']['mysql']['password'] = _in('MYSQL interface password','',use_default)

    clickhouse = _in('Connect to clickhouse ? [Y/N]','Y',use_default)
    if clickhouse in ['Y','y']:
        config['integrations']['clickhouse']['enabled'] = True
        config['integrations']['clickhouse']['host'] = _in('Clickhouse host: ','localhost',use_default)
        config['integrations']['clickhouse']['port'] = _in('Clickhouse port: ',8123,use_default)
        config['integrations']['clickhouse']['user'] = _in('Clickhouse user: ','default',use_default)
        config['integrations']['clickhouse']['password'] = _in('Clickhouse password: ','',use_default)

    mariadb = _in('Connect to Mariadb ? [Y/N]','Y',use_default)
    if clickhouse in ['Y','y']:
        config['integrations']['mariadb']['enabled'] = True
        config['integrations']['mariadb']['host'] = _in('Mariadb host: ','localhost',use_default)
        config['integrations']['mariadb']['port'] = _in('Mariadb port: ',3306,use_default)
        config['integrations']['mariadb']['user'] = _in('Mariadb user: ','root',use_default)
        config['integrations']['mariadb']['password'] = _in('Mariadb password: ','',use_default)

    config_path = os.path.join(config_dir,'config.json')
    with open(config_path, 'w') as fp:
        json.dump(config, fp, indent=4, sort_keys=True)

    return config_path


def daemon_creator(python_path,config_path):
    service_txt = f"""[Unit]
Description=Mindsdb

[Service]
ExecStart={python_path} -m mindsdb --config={config_path}

[Install]
WantedBy=multi-user.target
""".strip(' ')

    try:
        with open('/etc/systemd/system/mindsdb.service', 'w') as fp:
            fp.write(service_txt)
    except Exception as e:
        print(f'Failed to create daemon, error: {e}')

    try:
        os.system('systemctl daemon-reload')
    except Exception as e:
        print(f'Failed to load daemon, error: {e}')

def make_executable(python_path,config_path,path):
    text = f"""#!/bin/bash
{python_path} -m mindsdb --config={config_path}
"""

    with open(path, 'w') as fp:
        fp.write(text)

    os.system(f'chmod +x {path}')



import os
import logging
from logging.handlers import RotatingFileHandler

def init_logger(config, nolog=False):
    config = config['api']['mysql']['log']

    if not os.path.exists(config['folder']):
        os.makedirs(config['folder'])

    logger = logging.getLogger('mindsdb_sql')

    logger.setLevel(min(
        getattr(logging, config['file_level']),
        getattr(logging, config['console_level'])
    ))

    fh = RotatingFileHandler(
        os.path.join(config['folder'], config['file']),
        mode='a',
        encoding=config.get('encoding', 'utf-8'),
        maxBytes=100*1024,
        backupCount=3
    )
    fh.setLevel(config['file_level'])

    # create console handler
    ch = logging.StreamHandler()
    ch.setLevel(config['console_level'])

    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)

    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

log = logging.getLogger('mindsdb_sql')



import requests
from mindsdb.libs.constants.mindsdb import DATA_TYPES, DATA_SUBTYPES


class Clickhouse():
    def __init__(self, config):
        self.config = config
        self.host = config['integrations']['clickhouse']['host']
        self.port = config['integrations']['clickhouse']['port']
        self.user = config['integrations']['clickhouse']['user']
        self.password = config['integrations']['clickhouse']['password']

        self.setup_clickhouse()


    def _to_clickhouse_table(self, stats):
        subtype_map = {
            DATA_SUBTYPES.INT: 'Nullable(Int64)',
            DATA_SUBTYPES.FLOAT: 'Nullable(Float64)',
            DATA_SUBTYPES.BINARY: 'Nullable(UInt8)',
            DATA_SUBTYPES.DATE: 'Nullable(Date)',
            DATA_SUBTYPES.TIMESTAMP: 'Nullable(Datetime)',
            DATA_SUBTYPES.SINGLE: 'Nullable(String)',
            DATA_SUBTYPES.MULTIPLE: 'Nullable(String)',
            DATA_SUBTYPES.IMAGE: 'Nullable(String)',
            DATA_SUBTYPES.VIDEO: 'Nullable(String)',
            DATA_SUBTYPES.AUDIO: 'Nullable(String)',
            DATA_SUBTYPES.TEXT: 'Nullable(String)',
            DATA_SUBTYPES.ARRAY: 'Array(Float64)'
        }

        column_declaration = []
        for name, column in stats.items():
            try:
                col_subtype = stats[name]['typing']['data_subtype']
                new_type = subtype_map[col_subtype]
                column_declaration.append(f' `{name}` {new_type} ')
            except Exception as e:
                print(e)
                print(f'Error: cant convert type {col_subtype} of column {name} to clickhouse tpye')

        return column_declaration

    def _query(self, query):
        params = {'user': 'default'}
        try:
            params['user'] = self.config['integrations']['clickhouse']['user']
        except:
            pass

        try:
            params['password'] = self.config['integrations']['clickhouse']['password']
        except:
            pass

        host = self.config['integrations']['clickhouse']['host']
        port = self.config['integrations']['clickhouse']['port']

        response = requests.post(f'http://{host}:{port}', data=query, params=params)

        return response

    def setup_clickhouse(self):
        self._query('CREATE DATABASE IF NOT EXISTS mindsdb')

        msqyl_conn = self.config['api']['mysql']['host'] + ':' + str(self.config['api']['mysql']['port'])
        msqyl_user = self.config['api']['mysql']['user']
        msqyl_pass = self.config['api']['mysql']['password']

        q = f"""
                CREATE TABLE IF NOT EXISTS mindsdb.predictors
                (name String,
                status String,
                accuracy String,
                predict_cols String,
                select_data_query String,
                training_options String
                ) ENGINE=MySQL('{msqyl_conn}', 'mindsdb', 'predictors_clickhouse', '{msqyl_user}', '{msqyl_pass}')
        """
        print(f'Executing table creation query to create predictors list:\n{q}\n')
        self._query(q)

        q = f"""
            CREATE TABLE IF NOT EXISTS mindsdb.commands (
                command String
            ) ENGINE=MySQL('{msqyl_conn}', 'mindsdb', 'commands_clickhouse', '{msqyl_user}', '{msqyl_pass}')
        """
        print(f'Executing table creation query to create command table:\n{q}\n')
        self._query(q)

    def register_predictor(self, name, stats):
        columns_sql = ','.join(self._to_clickhouse_table(stats))
        columns_sql += ',`$select_data_query` Nullable(String)'

        msqyl_conn = self.config['api']['mysql']['host'] + ':' + str(self.config['api']['mysql']['port'])
        msqyl_user = self.config['api']['mysql']['user']
        msqyl_pass = self.config['api']['mysql']['password']

        q = f"""
                CREATE TABLE mindsdb.{name}
                ({columns_sql}
                ) ENGINE=MySQL('{msqyl_conn}', 'mindsdb', '{name}_clickhouse', '{msqyl_user}', '{msqyl_pass}')
        """
        print(f'Executing table creation query to sync predictor:\n{q}\n')
        self._query(q)

    def unregister_predictor(self, name):
        q = f"""
            drop table if exists mindsdb.{name};
        """
        print(f'Executing table creation query to sync predictor:\n{q}\n')
        self._query(q)



import os
import json


class Config(object):
    _config = {}

    def __init__(self, config='mindsdb/default_config.json'):
        if isinstance(config, dict):
            self._update_recursive(self._config, config)
        elif isinstance(config, str):
            self.merge(config)
        else:
            raise TypeError('argument must be string or dict')

    def merge(self, file_path):
        if os.path.isfile(file_path):
            with open(file_path, 'r') as fp:
                config = json.load(fp)
            self._update_recursive(self._config, config)

    def _update_recursive(self, a, b):
        for key in [x for x in a.keys() if x in b]:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                self._update_recursive(a[key], b[key])
            else:
                a[key] = b[key]
        for key in b.keys():
            if key not in a:
                a[key] = b[key]

    def __getitem__(self, key):
        return self._config[key]

    def get(self, key, default=None):
        return self._config.get(key, default)

    def get_all(self):
        return self._config



from flask import Flask, url_for
from flask_restx import Api
import json
from flask import g

from mindsdb.interfaces.datastore.datastore import DataStore
from mindsdb.interfaces.native.mindsdb import MindsdbNative


class Swagger_Api(Api):
    """
    This is a modification of the base Flask Restplus Api class due to the issue described here
    https://github.com/noirbizarre/flask-restplus/issues/223
    """
    @property
    def specs_url(self):
        return url_for(self.endpoint("specs"), _external=False)

def initialize_flask(config):
    app = Flask(__name__)

    app.config['SWAGGER_HOST'] = 'http://localhost:8000/mindsdb'
    authorizations = {
        'apikey': {
            'type': 'apiKey',
            'in': 'query',
            'name': 'apikey'
        }
    }

    api = Swagger_Api(app, authorizations=authorizations, security=['apikey'], url_prefix=':8000')

    return app, api

def initialize_interfaces(config,app):
    with app.app_context():
        g.default_store = DataStore(config)
        g.mindsdb_native = MindsdbNative(config)



import torch.multiprocessing as mp


ctx = mp.get_context('spawn')
class PredictorProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(PredictorProcess, self).__init__(args=args)

    def run(self):
        '''
        running at subprocess due to
        ValueError: signal only works in main thread

        this is work for celery worker here?
        '''
        import sys
        import mindsdb_native

        from mindsdb.utilities.config import Config

        name, from_data, to_predict, kwargs, config = self._args
        config = Config(config)

        mdb = mindsdb_native.Predictor(name=name)

        data_source = getattr(mindsdb, from_data['class'])(*from_data['args'], **from_data['kwargs'])

        mdb.learn(
            from_data=data_source,
            to_predict=to_predict,
            use_gpu=config.get('use_gpu', False),
            **kwargs
        )

        stats = mdb.get_model_data()['data_analysis_v2']

        try:
            assert(config['integrations']['clickhouse']['enabled'] == True)
            from mindsdb.interfaces.clickhouse.clickhouse import Clickhouse
            clickhouse = Clickhouse(config)
            clickhouse.register_predictor(name, stats)
        except:
            pass

        try:
            assert(config['integrations']['mariadb']['enabled'] == True)
            from mindsdb.interfaces.mariadb.mariadb import Mariadb
            mariadb = Mariadb(config)
            mariadb.register_predictor(name, stats)
        except:
            pass



import argparse

from mindsdb.api.mysql.mysql_proxy.mysql_proxy import MysqlProxy
from mindsdb.utilities.config import Config

def start(config):
    config = Config(config)
    MysqlProxy.startProxy(config)

if __name__ == '__main__':
    start()



# Mindsdb native interface
import mindsdb_native
from dateutil.parser import parse as parse_datetime

from mindsdb.interfaces.native.predictor_process import PredictorProcess


class MindsdbNative():
    def __init__(self, config):
        self.config = config
        self.metapredictor = mindsdb_native.Predictor('metapredictor')
        self.unregister_from = []

        try:
            assert(config['integrations']['clickhouse']['enabled'] == True)
            from mindsdb.interfaces.clickhouse.clickhouse import Clickhouse
            clickhouse = Clickhouse(self.config)
            self.unregister_from.append(clickhouse)
        except Exception as e:
            print(e)
            pass

        try:
            assert(config['integrations']['mariadb']['enabled'] == True)
            from mindsdb.interfaces.mariadb.mariadb import Mariadb
            mariadb = Mariadb(self.config)
            self.unregister_from.append(mariadb)
        except Exception as e:
            print(e)
            pass

    def learn(self, name, from_data, to_predict, kwargs={}):
        p = PredictorProcess(name, from_data, to_predict, kwargs, self.config.get_all())
        p.start()

    def predict(self, name, when=None, when_data=None, kwargs={}):
        mdb = mindsdb_native.Predictor(name=name)

        use_gpu = self.config.get('use_gpu', False)
        if when is not None:
            predictions = mdb.predict(
                when=when,
                run_confidence_variation_analysis=True,
                use_gpu=use_gpu,
                **kwargs
            )
        else:
            predictions = mdb.predict(
                when_data=when_data,
                run_confidence_variation_analysis=False,
                use_gpu=use_gpu,
                **kwargs
            )

        return predictions

    def analyse_dataset(self, ds):
        return self.metapredictor.analyse_dataset(ds, sample_margin_of_error=0.025)

    def get_model_data(self, name):
        return self.metapredictor.get_model_data(name)

    def get_models(self, status='any'):
        models = self.metapredictor.get_models()
        models = [x for x in models if x['name'] != 'metapredictor']
        if status != 'any':
            models = [x for x in models if x['status'] == status]

        for i in range(len(models)):
            for k in ['train_end_at', 'updated_at', 'created_at']:
                if k in models[i] and models[i][k] is not None:
                    try:
                        models[i][k] = parse_datetime(str(models[i][k]).split('.')[0])
                    except Exception as e:
                        models[i][k] = parse_datetime(str(models[i][k]))
        return models

    def delete_model(self, name):
        self.metapredictor.delete_model(name)
        for entity in self.unregister_from:
            unregister_func = getattr(entity, 'unregister_predictor')
            unregister_func(name)

    def rename_model(self, name, new_name):
        self.metapredictor.rename_model(name, new_name)

    def load_model(self, fpath):
        self.metapredictor.load_model(model_archive_path=fpath)

    def export_model(self,name):
        self.metapredictor.export_model(model_name=name)



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""


import random
import socketserver as SocketServer
import ssl
import re
import traceback
import json
import atexit

from moz_sql_parser import parse

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.api.mysql.mysql_proxy.controllers.log import init_logger, log
from mindsdb.api.mysql.mysql_proxy.datahub import init_datahub
from mindsdb.api.mysql.mysql_proxy.classes.client_capabilities import ClentCapabilities

from mindsdb.api.mysql.mysql_proxy.classes.sql_query import (
    SQLQuery,
    TableWithoutDatasourceException,
    UndefinedColumnTableException,
    DuplicateTableNameException,
    NotImplementedError,
    SqlError
)

from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import (
    getConstName,
    CHARSET_NUMBERS,
    ERR,
    COMMANDS,
    TYPES,
    SERVER_VARIABLES,
    DEFAULT_AUTH_METHOD
)

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets import (
    ErrPacket,
    HandshakePacket,
    FastAuthFail,
    PasswordAnswer,
    HandshakeResponsePacket,
    OkPacket,
    SwitchOutPacket,
    SwitchOutResponse,
    CommandPacket,
    ColumnCountPacket,
    ColumnDefenitionPacket,
    ResultsetRowPacket,
    EofPacket
)

from mindsdb.interfaces.datastore.datastore import DataStore
from mindsdb.interfaces.native.mindsdb import MindsdbNative


connection_id = 0

ALPHABET = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

HARDCODED_USER = None
HARDCODED_PASSWORD = None
CERT_PATH = None
default_store = None
mdb = None
datahub = None


class MysqlProxy(SocketServer.BaseRequestHandler):
    """
    The Main Server controller class
    """

    charset = 'utf8'

    charset_text_type = CHARSET_NUMBERS['utf8_general_ci']

    session = None

    client_capabilities = None

    @staticmethod
    def server_close(srv):
        srv.server_close()

    def initSession(self):
        global connection_id, ALPHABET
        log.info('New connection [{ip}:{port}]'.format(
            ip=self.client_address[0], port=self.client_address[1]))
        log.debug(self.__dict__)

        connection_id += 1

        self.session = SessionController()
        self.salt = ''.join([random.choice(ALPHABET) for i in range(20)])
        self.socket = self.request
        self.count = 0  # next packet number
        self.connection_id = connection_id
        self.logging = log

        self.current_transaction = None

        log.debug('session salt: {salt}'.format(salt=self.salt))

    def isAuthOk(self, user, orig_user, password, orig_password):
        try:
            log.info(f'check auth, user={user}')
            if user != orig_user:
                log.warning(f'check auth: user mismatch')
                return False
            if password != orig_password:
                log.warning(f'check auth: password mismatch')
                # @DEBUG === REMOVE !!
                return True
                #return False

            self.session.username = user
            self.session.auth = True
            log.info(f'auth ok, user={user}')
            return True
        except Exception:
            log.error(traceback.format_exc())
            log.error('failed to authenticate')

    def handshake(self):
        global HARDCODED_PASSWORD, HARDCODED_USER

        def switch_auth(method='mysql_native_password'):
            self.packet(SwitchOutPacket, seed=self.salt, method=method).send()
            switch_out_answer = self.packet(SwitchOutResponse)
            switch_out_answer.get()
            return switch_out_answer.enc_password.value

        if self.session is None:
            self.initSession()
        log.info('send HandshakePacket')
        self.packet(HandshakePacket).send()

        handshake_resp = self.packet(HandshakeResponsePacket)
        handshake_resp.get()
        if handshake_resp.length == 0:
            log.warning('HandshakeResponsePacket empty')
            self.packet(OkPacket).send()
            return False
        self.client_capabilities = ClentCapabilities(handshake_resp.capabilities.value)

        client_auth_plugin = handshake_resp.client_auth_plugin.value.decode()

        orig_username = HARDCODED_USER
        orig_password = HARDCODED_PASSWORD
        orig_password_hash = handshake_resp.scramble_func(HARDCODED_PASSWORD, self.salt)
        username = None
        password = None

        if handshake_resp.type == 'SSLRequest':
            ssl_socket = ssl.wrap_socket(
                self.socket,
                server_side=True,
                certfile=CERT_PATH,
                do_handshake_on_connect=True
            )
            self.socket = ssl_socket
            handshake_resp = self.packet(HandshakeResponsePacket)
            handshake_resp.get()
            client_auth_plugin = handshake_resp.client_auth_plugin.value.decode()

            if DEFAULT_AUTH_METHOD not in client_auth_plugin:
                # if methods mismatch - need to switch them
                new_method = 'caching_sha2_password' if 'caching_sha2_password' in client_auth_plugin else 'mysql_native_password'
                password = switch_auth(new_method)

                if new_method == 'caching_sha2_password':
                    self.packet(FastAuthFail).send()
                    password_answer = self.packet(PasswordAnswer)
                    password_answer.get()
                    password = password_answer.password.value.decode()
                else:
                    orig_password = orig_password_hash

            elif 'caching_sha2_password' in client_auth_plugin:
                self.packet(FastAuthFail).send()
                log.info('connected with SSL, caching_sha2_password method')
                password_answer = self.packet(PasswordAnswer)
                password_answer.get()
                password = password_answer.password.value.decode()
                orig_password = HARDCODED_PASSWORD
            elif 'mysql_native_password' in client_auth_plugin:
                log.info('connected with SSL, mysql_native_password method')
                password = handshake_resp.enc_password.value
                orig_password = orig_password_hash
            else:
                log.info('connected with SSL, unknown method, try switch to mysql_native_password')
                password = switch_auth('mysql_native_password')
                orig_password = orig_password_hash
        else:
            if DEFAULT_AUTH_METHOD not in client_auth_plugin:
                new_method = 'caching_sha2_password' if 'caching_sha2_password' in client_auth_plugin else 'mysql_native_password'
                if new_method == 'caching_sha2_password':
                    self.packet(ErrPacket, err_code=ERR.ER_PASSWORD_NO_MATCH, msg=f'caching_sha2_password without SSL not supported').send()
                    return False
                password = switch_auth(new_method)
                orig_password = orig_password_hash

            elif 'mysql_native_password' in client_auth_plugin:
                log.info('connected without SSL, default method')
                password = handshake_resp.enc_password.value
                orig_password = orig_password_hash
            else:
                log.info('connected without SSL, unknown method, try switch to mysql_native_password')
                password = switch_auth('mysql_native_password')
                orig_password = orig_password_hash

        username = handshake_resp.username.value.decode()

        try:
            self.session.database = handshake_resp.database.value.decode()
        except Exception:
            self.session.database = None

        if self.isAuthOk(username, orig_username, password, orig_password):
            self.packet(OkPacket).send()
            return True
        else:
            self.packet(ErrPacket, err_code=ERR.ER_PASSWORD_NO_MATCH, msg=f'Access denied for user {username}').send()
            log.warning('AUTH FAIL')
            return False

    def sendPackageGroup(self, packages):
        string = b''.join([x.accum() for x in packages])
        self.socket.sendall(string)

    def answerVersionComment(self):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'table_name': '',
                'name': '@@version_comment',
                'type': TYPES.MYSQL_TYPE_VAR_STRING
            }],
            data=[{'@@version_comment': '(MindsDB)'}]
        )
        if self.client_capabilities.DEPRECATE_EOF is True:
            packages.append(self.packet(OkPacket, eof=True))
        else:
            packages.append(self.packet(EofPacket))
        self.sendPackageGroup(packages)

    def answerVersion(self):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'table_name': '',
                'name': '@@version',
                'type': TYPES.MYSQL_TYPE_VAR_STRING
            }],
            data=[{'@@version': '0.1'}]
        )
        packages.append(self.packet(OkPacket, eof=True))
        self.sendPackageGroup(packages)

    def answerTableQuery(self, query):
        packages = []
        packages += self.getTabelPackets(
            columns=query.columns,
            data=query.result
        )
        packages.append(self.packet(OkPacket, eof=True))
        self.sendPackageGroup(packages)

    def insert_predictor_answer(self, sql):
        insert = SQLQuery.parse_insert(sql)

        datasources = default_store.get_datasources()
        if insert['name'] in [x['name'] for x in datasources]:
            self.packet(ErrPacket, err_code=ERR.ER_WRONG_ARGUMENTS, msg=f"datasource with name '{insert['name']}'' already exists").send()
            return

        kwargs = {}
        if isinstance(insert.get('training_options'), str) \
            and len(insert['training_options']) > 0:
            try:
                kwargs = json.loads(insert['training_options'])
            except Exception as e:
                self.packet(
                    ErrPacket,
                    err_code=ERR.ER_WRONG_ARGUMENTS,
                    msg='training_options should be in valid JSON string'
                ).send()
                return

        # TODO clickhouse with any type of used escaping sends escaped quotes as \'.
        # Need to check other clients, they behaviour can be differ
        insert['select_data_query'] = insert['select_data_query'].replace(r"\'", "'")

        db = sql.lower()[sql.lower().find('predictors_') + len('predictors_'):]
        db = db[:db.find(' ')].strip(' `')
        ds_type = db
        ds = default_store.save_datasource(insert['name'], ds_type, insert['select_data_query'])
        mdb.learn(insert['name'], ds, insert['predict_cols'], kwargs)

        self.packet(OkPacket).send()

    def delete_predictor_answer(self, sql):
        global datahub

        fake_sql = sql.strip(' ')
        fake_sql = 'select name ' + fake_sql[len('delete '):]
        query = SQLQuery(fake_sql)

        result = query.fetch(datahub)

        if result['success'] is False:
            self.packet(
                ErrPacket,
                err_code=result['error_code'],
                msg=result['msg']
            ).send()
            return

        predictors_names = [x[0] for x in result['result']]

        if len(predictors_names) == 0:
            raise NotImplementedError('nothing to delete')

        for predictor_name in predictors_names:
            datahub['mindsdb'].delete_predictor(predictor_name)

        self.packet(OkPacket).send()

    def handle_custom_command(self, sql):
        insert = SQLQuery.parse_insert(sql)

        if 'command' not in insert:
            self.packet(ErrPacket, err_code=ERR.ER_WRONG_ARGUMENTS, msg=f"command should be inserted").send()
            return
        if len(insert) > 1:
            self.packet(ErrPacket, err_code=ERR.ER_WRONG_ARGUMENTS, msg=f"only command should be inserted").send()
            return

        command = insert['command'].strip(' ;').split()

        if command[0].lower() == 'delete' and command[1].lower() == 'predictor':
            if len(command) != 3:
                self.packet(
                    ErrPacket,
                    err_code=ERR.ER_SYNTAX_ERROR,
                    msg="wrong syntax of 'DELETE PREDICTOR {NAME}' command"
                ).send()
                return
            predictor_name = command[2]
            self.delete_predictor_answer(f"delete from mindsdb.predictors where name = '{predictor_name}'")
            return

        self.packet(
            ErrPacket,
            err_code=ERR.ER_SYNTAX_ERROR,
            msg="at this moment only 'delete predictor' command supported"
        ).send()

    def queryAnswer(self, sql):
        sql_lower = sql.lower()
        sql_lower = sql_lower.replace('`', '')

        if 'show databases' in sql_lower:
            sql = 'select schema_name as Database from information_schema.SCHEMATA;'
            sql_lower = sql.lower()
        if 'show full tables from' in sql_lower:
            schema = re.findall(r'show\s+full\s+tables\s+from\s+(\S*)', sql_lower)[0]
            sql = f"select table_name as Tables_in_{schema} from INFORMATION_SCHEMA.TABLES WHERE table_schema = '{schema.upper()}' and table_type = 'BASE TABLE'"
            sql_lower = sql.lower()

        keyword = sql_lower.split(' ')[0]

        if keyword == 'start':
            # start transaction
            self.packet(OkPacket).send()
            return
        elif keyword == 'set':
            if 'autocommit' in sql_lower:
                self.packet(OkPacket).send()
                return
            elif 'set names' in sql_lower:
                # it can be "set names utf8"
                self.charset = re.findall(r"set\s+names\s+(\S*)", sql_lower)[0]
                self.charset_text_type = CHARSET_NUMBERS['utf8_general_ci']
                if self.charset == 'utf8mb4':
                    self.charset_text_type = CHARSET_NUMBERS['utf8mb4_general_ci']
                self.packet(
                    OkPacket,
                    state_track=[
                        ['character_set_client', self.charset],
                        ['character_set_connection', self.charset],
                        ['character_set_results', self.charset]
                    ]
                ).send()
                return
            else:
                self.packet(OkPacket).send()
                return
        elif keyword == 'use':
            self.session.database = sql_lower.split()[1].trim(' ;')
            self.packet(OkPacket).send()
            return
        elif 'show warnings' in sql_lower:
            self.answerShowWarnings()
            return
        elif 'show engines' in sql_lower:
            self.answerShowEngines()
            return
        elif 'show charset' in sql_lower:
            self.answerShowCharset()
            return
        elif 'show collation' in sql_lower:
            self.answerShowCollation()
            return
        elif keyword == 'delete' and \
            ('mindsdb.predictors' in sql_lower or self.session.database == 'mindsdb' and 'predictors' in sql_lower):
            self.delete_predictor_answer(sql, db)
            return
        elif keyword == 'insert' and \
            ('mindsdb.commands' in sql_lower or self.session.database == 'mindsdb' and 'commands' in sql_lower):
            self.handle_custom_command(sql)
            return
        elif keyword == 'insert' and \
            ('mindsdb.predictors' in sql_lower or self.session.database == 'mindsdb' and 'predictors' in sql_lower):
            self.insert_predictor_answer(sql)
            return
        elif keyword in ('update', 'insert'):
            raise NotImplementedError('Update and Insert not implemented')
            return
        elif keyword == 'alter' and ('disable keys' in sql_lower) or ('enable keys' in sql_lower):
            self.packet(OkPacket).send()
            return
        elif keyword == 'select':
            if '@@' in sql_lower:
                self.answerVariables(sql)
                return
            if 'select 1' in sql_lower:
                self.answerSelect1(sql)
                return
            if 'database()' in sql_lower:
                self.answerSelectDatabase()
                return

            # TODO rewrite it
            # if '`predictors_mariadb`' in sql:
            #     sql = sql.replace('`predictors_mariadb`', '`mindsdb`.`predictors`')

            query = SQLQuery(sql, self.session.database)
            return self.selectAnswer(query)
        elif keyword == 'rollback':
            self.packet(OkPacket).send()
            return
        elif keyword == 'commit':
            self.packet(OkPacket).send()
            return
        else:
            raise NotImplementedError('Action not implemented')
            return

    def answerShowWarnings(self):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'database': '',
                'table_name': '',
                'name': 'Level',
                'alias': 'Level',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': '',
                'table_name': '',
                'name': 'Code',
                'alias': 'Code',
                'type': TYPES.MYSQL_TYPE_LONG,
                'charset': CHARSET_NUMBERS['binary']
            }, {
                'database': '',
                'table_name': '',
                'name': 'Message',
                'alias': 'Message',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }],
            data=[]
        )
        packages.append(self.packet(OkPacket, eof=True, status=0x0002))
        self.sendPackageGroup(packages)

    def answerSelectDatabase(self):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'database': '',
                'table_name': '',
                'name': 'database()',
                'alias': 'database()',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }],
            data=[
                [None]
            ]
        )
        packages.append(self.packet(OkPacket, eof=True, status=0x0000))
        self.sendPackageGroup(packages)

    def answerShowCollation(self):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'database': 'information_schema',
                'table_name': 'COLLATIONS',
                'name': 'COLLATION_NAME',
                'alias': 'Collation',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'COLLATIONS',
                'name': 'CHARACTER_SET_NAME',
                'alias': 'Charset',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'COLLATIONS',
                'name': 'ID',
                'alias': 'Id',
                'type': TYPES.MYSQL_TYPE_LONGLONG,
                'charset': CHARSET_NUMBERS['binary']
            }, {
                'database': 'information_schema',
                'table_name': 'COLLATIONS',
                'name': 'IS_DEFAULT',
                'alias': 'Default',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'COLLATIONS',
                'name': 'IS_COMPILED',
                'alias': 'Compiled',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'COLLATIONS',
                'name': 'SORTLEN',
                'alias': 'Sortlen',
                'type': TYPES.MYSQL_TYPE_LONGLONG,
                'charset': CHARSET_NUMBERS['binary']
            }],
            data=[
                ['utf8_general_ci', 'utf8', 33, 'Yes', 'Yes', 1],
                ['latin1_swedish_ci', 'latin1', 8, 'Yes', 'Yes', 1]
            ]
        )
        packages.append(self.packet(OkPacket, eof=True, status=0x0002))
        self.sendPackageGroup(packages)

    def answerShowCharset(self):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'database': 'information_schema',
                'table_name': 'CHARACTER_SETS',
                'name': 'CHARACTER_SET_NAME',
                'alias': 'Charset',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'CHARACTER_SETS',
                'name': 'DESCRIPTION',
                'alias': 'Description',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'CHARACTER_SETS',
                'name': 'DEFAULT_COLLATE_NAME',
                'alias': 'Default collation',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'CHARACTER_SETS',
                'name': 'MAXLEN',
                'alias': 'Maxlen',
                'type': TYPES.MYSQL_TYPE_LONGLONG,
                'charset': CHARSET_NUMBERS['binary']
            }],
            data=[
                ['utf8', 'UTF-8 Unicode', 'utf8_general_ci', 3],
                ['latin1', 'cp1252 West European', 'latin1_swedish_ci', 1]
            ]
        )
        packages.append(self.packet(OkPacket, eof=True, status=0x0002))
        self.sendPackageGroup(packages)

    def answerShowEngines(self):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'database': 'information_schema',
                'table_name': 'ENGINES',
                'name': 'Engine',
                'alias': 'Engine',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'ENGINES',
                'name': 'Support',
                'alias': 'Support',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'ENGINES',
                'name': 'Comment',
                'alias': 'Comment',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'ENGINES',
                'name': 'Transactions',
                'alias': 'Transactions',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'ENGINES',
                'name': 'XA',
                'alias': 'XA',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }, {
                'database': 'information_schema',
                'table_name': 'ENGINES',
                'name': 'Savepoints',
                'alias': 'Savepoints',
                'type': TYPES.MYSQL_TYPE_VAR_STRING,
                'charset': self.charset_text_type
            }],
            data=[['InnoDB', 'DEFAULT', 'Supports transactions, row-level locking, and foreign keys', 'YES', 'YES', 'YES']]
        )
        packages.append(self.packet(OkPacket, eof=True, status=0x0002))
        self.sendPackageGroup(packages)

    def answerSelect1(self, sql):
        packages = []
        packages += self.getTabelPackets(
            columns=[{
                'table_name': '',
                'name': '',
                'alias': '1',
                'type': TYPES.MYSQL_TYPE_LONGLONG,
                'charset': CHARSET_NUMBERS['binary']
            }],
            data=[[1]]
        )
        packages.append(self.packet(OkPacket, eof=True, status=0x0002))
        self.sendPackageGroup(packages)

    def answerVariables(self, sql):
        if '@@version_comment' in sql.lower():
            self.answerVersionComment()
            return
        elif '@@version' in sql.lower():
            self.answerVersion()
            return
        p = parse(sql)
        select = p.get('select')
        if isinstance(select, dict):
            select = [select]
        columns = []
        row = []
        for s in select:
            fill_name = s.get('value')
            alias = s.get('name', fill_name)
            variable = SERVER_VARIABLES.get(fill_name)
            columns.append({
                'table_name': '',
                'name': '',
                'alias': alias,
                'type': variable[1],
                'charset': variable[2]
            })
            row.append(variable[0])

        packages = []
        packages += self.getTabelPackets(
            columns=columns,
            data=[row]
        )
        packages.append(self.packet(OkPacket, eof=True, status=0x0002))
        self.sendPackageGroup(packages)

    def selectAnswer(self, query):
        global datahub
        result = query.fetch(datahub)

        if result['success'] is False:
            self.packet(
                ErrPacket,
                err_code=result['error_code'],
                msg=result['msg']
            ).send()
            return

        self.answerTableQuery(query)

    def getTabelPackets(self, columns, data):
        # TODO remove columns order
        packets = [self.packet(ColumnCountPacket, count=len(columns))]
        for i, column in enumerate(columns):
            table_name = column.get('table_name', 'table_name')
            column_name = column.get('name', 'column_name')
            column_alias = column.get('alias', column_name)
            length = 1
            for row in data:
                if isinstance(row, dict):
                    length = max(len(str(row[column_alias])), length)
                else:
                    length = max(len(str(row[i])), length)
            packets.append(
                self.packet(
                    ColumnDefenitionPacket,
                    schema=column.get('database', 'mindsdb_schema'),
                    table_alias=column.get('table_alias', table_name),
                    table_name=table_name,
                    column_alias=column_alias,
                    column_name=column_name,
                    column_type=column['type'],
                    charset=column.get('charset', CHARSET_NUMBERS["utf8_unicode_ci"]),
                    max_length=length
                )
            )

        if self.client_capabilities.DEPRECATE_EOF is False:
            packets.append(self.packet(EofPacket))

        packets += [self.packet(ResultsetRowPacket, data=x) for x in data]
        return packets

    def handle(self):
        """
        Handle new incoming connections
        :return:
        """
        log.info('handle new incoming connection')
        if self.handshake() is False:
            return

        while True:
            log.debug('Got a new packet')
            p = self.packet(CommandPacket)

            try:
                success = p.get()
            except Exception:
                log.warning('Session closed, on packet read error')
                log.debug(traceback.format_exc())
                # self.server.shutdown()
                return

            if success is False:
                log.info('Session closed by client')
                # self.server.shutdown()
                return

            log.info('Command TYPE: {type}'.format(
                type=getConstName(COMMANDS, p.type.value)))

            if p.type.value == COMMANDS.COM_QUERY:
                try:
                    sql = p.sql.value.decode('utf-8')
                    # NOTE dbeaver can insert in start of query comment with connector version, for example
                    # /* mysql-connector-java-8.0.11 (Revision: 6d4eaa273bc181b4cf1c8ad0821a2227f116fedf) */SELECT @@session.auto_increment_increment
                    sql = re.sub(re.compile("/\*.*?\*/", re.DOTALL), "", sql)
                    sql = sql.strip(' ;')
                except Exception:
                    log.error('SQL contains non utf-8 values: {sql}'.format(sql=p.sql.value))
                    self.packet(OkPacket).send()
                    continue
                log.info(f'COM_QUERY: {sql}')
                self.current_transaction = self.session.newTransaction(sql_query=sql)

                try:
                    self.queryAnswer(sql)
                except Exception as e:
                    log.error(
                        f'ERROR while executing query: {sql}\n'
                        f'{traceback.format_exc()}\n'
                        f'{e}'
                    )
                    self.packet(
                        ErrPacket,
                        err_code=ERR.ER_SYNTAX_ERROR,
                        msg=str(e)
                    ).send()

                # if self.current_transaction.output_data_array is None:
                #     self.packet(OkPacket).send()
                # else:
                #     self.packet(ResultsetPacket, metadata=self.current_transaction.output_metadata,
                #                 data_array=self.current_transaction.output_data_array).send()
            elif p.type.value == COMMANDS.COM_QUIT:
                log.info('Session closed, on client disconnect')
                self.session = None
                break
            else:
                log.info('Command has no specific handler, return OK msg')
                log.debug(str(p))
                # p.pprintPacket() TODO: Make a version of print packet
                # that sends it to debug isntead
                self.packet(OkPacket).send()

    def packet(self, packetClass=Packet, **kwargs):
        """
        Factory method for packets

        :param packetClass:
        :param kwargs:
        :return:
        """
        p = packetClass(socket=self.socket, seq=self.count, session=self.session, proxy=self, **kwargs)
        self.count += 1
        return p

    @staticmethod
    def startProxy(config):
        global HARDCODED_USER
        global HARDCODED_PASSWORD
        global CERT_PATH
        global default_store
        global mdb
        global datahub
        """
        Create a server and wait for incoming connections until Ctrl-C
        """
        init_logger(config)

        HARDCODED_USER = config['api']['mysql']['user']
        HARDCODED_PASSWORD = config['api']['mysql']['password']
        CERT_PATH = config['api']['mysql']['certificate_path']
        default_store = DataStore(config)
        mdb = MindsdbNative(config)
        datahub = init_datahub(config)

        host = config['api']['mysql']['host']
        port = int(config['api']['mysql']['port'])

        log.info(f'Starting MindsDB Mysql proxy server on tcp://{host}:{port}')

        # Create the server
        if config['debug'] is True:
            SocketServer.TCPServer.allow_reuse_address = True
        server = SocketServer.ThreadingTCPServer((host, port), MysqlProxy)

        atexit.register(MysqlProxy.server_close, srv=server)

        # Activate the server; this will keep running until you
        # interrupt the program with Ctrl-C
        log.info('Waiting for incoming connections...')
        server.serve_forever()


if __name__ == "__main__":
    MysqlProxy.startProxy()



import sqlite3
from mindsdb.libs.constants.mindsdb import DATA_TYPES, DATA_SUBTYPES


def create_sqlite_db(path, data_frame):
    con = sqlite3.connect(path)
    data_frame.to_sql(name='data', con=con, index=False)
    con.close()

def cast_df_columns_types(df, stats):
    types_map = {
        DATA_TYPES.NUMERIC: {
            DATA_SUBTYPES.INT: 'int64',
            DATA_SUBTYPES.FLOAT: 'float64',
            DATA_SUBTYPES.BINARY: 'bool'
        },
        DATA_TYPES.DATE: {
            DATA_SUBTYPES.DATE: 'datetime64',       # YYYY-MM-DD
            DATA_SUBTYPES.TIMESTAMP: 'datetime64'   # YYYY-MM-DD hh:mm:ss or 1852362464
        },
        DATA_TYPES.CATEGORICAL: {
            DATA_SUBTYPES.SINGLE: 'category',
            DATA_SUBTYPES.MULTIPLE: 'category'
        },
        DATA_TYPES.FILE_PATH: {
            DATA_SUBTYPES.IMAGE: 'object',
            DATA_SUBTYPES.VIDEO: 'object',
            DATA_SUBTYPES.AUDIO: 'object'
        },
        DATA_TYPES.SEQUENTIAL: {
            DATA_SUBTYPES.TEXT: 'object',
            DATA_SUBTYPES.ARRAY: 'object'
        }
    }

    columns = [dict(name=x) for x in list(df.keys())]

    for column in columns:
        try:
            name = column['name']
            col_type = stats[name]['typing']['data_type']
            col_subtype = stats[name]['typing']['data_subtype']
            new_type = types_map[col_type][col_subtype]
            if new_type == 'int64' or new_type == 'float64':
                df[name] = df[name].apply(lambda x: x.replace(',','.') if isinstance(x, str) else x)
            if new_type == 'int64':
                df = df.astype({name: 'float64'})
            df = df.astype({name: new_type})
        except Exception as e:
            print(e)
            print(f'Error: cant convert type of DS column {name} to {new_type}')

    return df

def parse_filter(key, value):
    result = re.search(r'filter(_*.*)\[(.*)\]', key)
    operator = result.groups()[0].strip('_') or 'like'
    field = result.groups()[1]
    operators_map = {
        'like': 'like',
        'in': 'in',
        'nin': 'not in',
        'gt': '>',
        'lt': '<',
        'gte': '>=',
        'lte': '<=',
        'eq': '=',
        'neq': '!='
    }
    if operator not in operators_map:
        return None
    operator = operators_map[operator]
    return {'field': field, 'value': value, 'operator': operator}


def prepare_sql_where(where):
    marks = {}
    if len(where) > 0:
        for i in range(len(where)):
            field = where[i]['field'].replace('"', '""')
            operator = where[i]['operator']
            value = where[i]['value']
            var_name = f'var{i}'
            if ' ' in field:
                field = f'"{field}"'
            if operator == 'like':
                marks[var_name] = '%' + value + '%'
            else:
                marks[var_name] = value
            where[i] = f'{field} {operator} :var{i}'
        where = 'where ' + ' and '.join(where)
    else:
        where = ''
    return where, marks

def get_sqlite_columns_names(cursor):
    cursor.execute('pragma table_info(data);')
    column_name_index = [x[0] for x in cursor.description].index('name')
    columns = cursor.fetchall()
    return [x[column_name_index] for x in columns]

def get_sqlite_data(db_path, where, limit, offset):
    where = [] if where is None else where

    con = sqlite3.connect(db_path)
    cur = con.cursor()

    offset = '' if limit is None or offset is None else f'offset {offset}'
    limit = '' if limit is None else f'limit {limit}'

    columns_names = get_sqlite_columns_names(cur)
    where = [x for x in where if x['field'] in columns_names]
    where, marks = prepare_sql_where(where)

    count_query = ' '.join(['select count(1) from data', where])
    cur.execute(count_query, marks)
    rowcount = cur.fetchone()[0]

    query = ' '.join(['select * from data', where, limit, offset])
    cur.execute(query, marks)
    data = cur.fetchall()
    data = [dict(zip(columns_names, x)) for x in data]

    cur.close()
    con.close()

    return {
        'data': data,
        'rowcount': rowcount,
        'columns_names': columns_names
    }



import json
import datetime
from dateutil.parser import parse as parse_dt
import shutil
import os
import pickle
import sys
import resource

import mindsdb

from mindsdb.interfaces.datastore.sqlite_helpers import *
from mindsdb.interfaces.native.mindsdb import MindsdbNative
from mindsdb import FileDS, ClickhouseDS, MariaDS
from mindsdb.interfaces.datastore.sqlite_helpers import create_sqlite_db


class DataStore():
    def __init__(self, config, storage_dir=None):
        self.config = config
        self.dir = storage_dir if isinstance(storage_dir, str) else config['interface']['datastore']['storage_dir']
        self.mindsdb_native = MindsdbNative(config)

    def get_analysis(self, ds):
        return self.mindsdb_native.analyse_dataset(ds)

    def get_datasources(self):
        datasource_arr = []
        for ds_name in os.listdir(self.dir):
            try:
                with open(os.path.join(self.dir, ds_name, 'datasource', 'metadata.json'), 'r') as fp:
                    try:
                        datasource = json.load(fp)
                        datasource['created_at'] = parse_dt(datasource['created_at'].split('.')[0])
                        datasource['updated_at'] = parse_dt(datasource['updated_at'].split('.')[0])
                        datasource_arr.append(datasource)
                    except Exception as e:
                        print(e)
            except Exception as e:
                print(e)
        return datasource_arr

    def get_data(self, name, where=None, limit=None, offset=None):
        # @TODO Apply filter directly to postgres/mysql/clickhouse/etc...  when the datasource is of that type
        return get_sqlite_data(os.path.join(self.dir, name, 'datasource', 'sqlite.db'), where=where, limit=limit, offset=offset)

    def get_datasource(self, name):
        for ds in self.get_datasources():
            if ds['name'] == name:
                return ds
        return None

    def delete_datasource(self, name):
        data_sources = self.get_datasource(name)
        shutil.rmtree(os.path.join(self.dir, data_sources['name']))

    def save_datasource(self, name, source_type, source, file_path=None):
        print(name, source_type, source)
        if source_type == 'file' and (file_path is None):
            raise Exception('`file_path` argument required when source_type == "file"')

        for i in range(1, 1000):
            if name in [x['name'] for x in self.get_datasources()]:
                previous_index = i - 1
                name = name.replace(f'__{previous_index}__', '')
                name = f'{name}__{i}__'
            else:
                break

        ds_meta_dir = os.path.join(self.dir, name)
        os.mkdir(ds_meta_dir)

        ds_dir = os.path.join(ds_meta_dir, 'datasource')
        os.mkdir(ds_dir)

        print(source_type)
        if source_type == 'file':
            source = os.path.join(ds_dir, source)
            os.replace(file_path, source)
            ds = FileDS(source)
            picklable = {
                'class': 'FileDS'
                ,'args': [source]
                ,'kwargs': {}
            }
        elif source_type == 'clickhouse':
            user = self.config['integrations']['clickhouse']['user']
            password = self.config['integrations']['clickhouse']['password']
            # TODO add host port params
            ds = ClickhouseDS(source, user=user, password=password)
            picklable = {
                'class': 'ClickhouseDS'
                ,'args': [source]
                ,'kwargs': {'user': user,'password': password}
            }
        elif source_type == 'mariadb':
            user = self.config['integrations']['mariadb']['user']
            password = self.config['integrations']['mariadb']['password']
            host = self.config['integrations']['mariadb']['host']
            port = self.config['integrations']['mariadb']['port']
            ds = MariaDS(source, user=user, password=password, host=host, port=port)
            picklable = {
                'class': 'MariaDS'
                ,'args': [source]
                ,'kwargs': {
                    'user': user,
                    'password': password,
                    'host': host,
                    'port': port
                }
            }
        else:
            # This probably only happens for urls
            print('Create URL data source !')
            ds = FileDS(source)
            picklable = {
                'class': 'FileDS'
                ,'args': [source]
                ,'kwargs': {}
            }

        df = ds.df

        df_with_types = cast_df_columns_types(df, self.get_analysis(df)['data_analysis_v2'])
        create_sqlite_db(os.path.join(ds_dir, 'sqlite.db'), df_with_types)

        print(picklable)
        with open(os.path.join(ds_dir,'ds.pickle'), 'wb') as fp:
            pickle.dump(picklable, fp)

        with open(os.path.join(ds_dir,'metadata.json'), 'w') as fp:
            json.dump({
                'name': name,
                'source_type': source_type,
                'source': source,
                'created_at': str(datetime.datetime.now()).split('.')[0],
                'updated_at': str(datetime.datetime.now()).split('.')[0],
                'row_count': len(df),
                'columns': [dict(name=x) for x in list(df.keys())]
            }, fp)

        return self.get_datasource_obj(name)

    def get_datasource_obj(self, name):
        ds_meta_dir = os.path.join(self.dir, name)
        ds_dir = os.path.join(ds_meta_dir, 'datasource')
        try:
            #resource.setrlimit(resource.RLIMIT_STACK, [0x10000000, resource.RLIM_INFINITY])
            #sys.setrecursionlimit(0x100000)
            with open(os.path.join(ds_dir,'ds.pickle'), 'rb') as fp:
                picklable = pickle.load(fp)
                #ds = picklable['class'](*picklable['args'],**picklable['kwargs'])

            return picklable
        except Exception as e:
            print('\n\n\n')
            print(e)
            print('\n\n\n')
            return None



import argparse
import atexit
from torch.multiprocessing import Process
import traceback
import sys
import os

from mindsdb.utilities.config import Config
from mindsdb.api.http.start import start as start_http
from mindsdb.api.mysql.start import start as start_mysql
from mindsdb.utilities.fs import get_or_create_dir_struct


def close_api_gracefully(p_arr):
    for p in p_arr:
        sys.stdout.flush()
        p.terminate()
        p.join()
        sys.stdout.flush()


parser = argparse.ArgumentParser(description='CL argument for mindsdb server')
parser.add_argument('--api', type=str, default=None)
parser.add_argument('--config', type=str, default=None)

args = parser.parse_args()

config_path = args.config
if config_path is None:
    config_dir, _, _ = get_or_create_dir_struct()
    config_path = os.path.join(config_dir,'config.json')

print(f'Using configuration file: {config_path}')
config = Config(config_path)

if args.api is None:
    api_arr = [api for api in config['api']]
else:
    api_arr = args.api.split(',')

start_functions = {
    'http': start_http,
    'mysql': start_mysql
}

p_arr = []
for api in api_arr:
    print(f'Starting Mindsdb {api} API !')
    try:
        p = Process(target=start_functions[api], args=(config_path,))
        p.start()
        p_arr.append(p)
        print(f'Started Mindsdb {api} API !')
    except Exception as e:
        close_api_gracefully(p_arr)
        print(f'Failed to start {api} API with exception {e}')
        print(traceback.format_exc())
        raise

atexit.register(close_api_gracefully, p_arr=p_arr)

for p in p_arr:
    p.join()



import os
import json

from mindsdb.utilities.fs import get_or_create_dir_struct
from mindsdb.utilities.wizards import cli_config

config_dir, predictor_dir, datasource_dir = get_or_create_dir_struct()
config_path = os.path.join(config_dir,'config.json')
if not os.path.exists(config_path):
    _ = cli_config(None,None,predictor_dir,datasource_dir,config_dir,use_default=True)

with open(config_path, 'r') as fp:
    os.environ['MINDSDB_STORAGE_PATH'] = json.load(fp)['interface']['mindsdb_native']['storage_dir']


from mindsdb_native import *
# Figure out how to add this as a module
import lightwood
#import dataskillet
import mindsdb.utilities.wizards as wizards



from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from mindsdb.api.http.namespaces.entitites.column_metadata import column_metadata

from flask_restx import fields

data_analysis_metadata = ns_conf.model('PredictorDataAnalysisMetadata', {
    'target_columns_metadata': fields.List( fields.Nested(column_metadata), required=False, description='The number of rows used on the validation subset'),
    'input_columns_metadata': fields.List( fields.Nested(column_metadata), required=False, description='The number of rows used on the validation subset')
})





"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.controllers.log import log


class SessionController():
    '''
    This class manages the server session
    '''

    def __init__(self) -> object:
        """
        Initialize the session
        :param socket:
        """

        self.username = None
        self.auth = False
        self.logging = log

        self.current_transaction = None

    def newTransaction(self, sql_query):
        return {
            'output_metadata': {},
            'output_data_array': []
        }



from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from mindsdb.api.http.namespaces.entitites.data_preparation_metadata import data_preparation_metadata, EXAMPLE as PREPARATION_METADATA_EXAMPLE
from mindsdb.api.http.namespaces.entitites.data_analysis_metadata import data_analysis_metadata#, EXAMPLE as DATA_ANALYSIS_METADATA_EXAMPLE
from mindsdb.api.http.namespaces.entitites.target_column_metadata import target_column_metadata#, EXAMPLES as TARGET_COLUMN_METADATA_EXAMPLES
from flask_restx import fields
from collections import OrderedDict

predictor_metadata = ns_conf.model('PredictorMetadata', {
    # Primary key
    'status': fields.String(required=False, description='The current model status', enum=['training', 'complete', 'error']),
    'current_phase': fields.String(required=False, description='Current training phase'),
    'name': fields.String(required=False, description='The predictor name'),
    'version': fields.String(required=False, description='The predictor version to publish under, this is so that we can train multiple predictors for the same problem but expose them via the same name'),
    # other attributes
    'data_preparation': fields.Nested(data_preparation_metadata, required=False, description='The metadata used in the preparation stage, in which we break the data into train, test, validation'),
    'accuracy': fields.Float(description='The current accuracy of the model'),
    'data_analysis': fields.Nested(data_analysis_metadata, required=False, description='The metadata used in the analysis stage, in which we extract statistical information from the input data'),
    'model_analysis': fields.List(fields.Nested(target_column_metadata), required=False, description='The model analysis stage, in which we extract statistical information from the input data for each target variable, thus, this is a list; one item per target column')
    ,'data_analysis_v2': fields.Raw(default={})
})

predictor_query_params = OrderedDict([
    ('name', {
        'description': 'The predictor name',
        'type': 'string',
        'in': 'path',
        'required': True
    }),
    ('when', {
        'description': 'The query conditions: key - column, value - value',
        'type': 'object',
        'in': 'body',
        'required': True,
        'example': "{'number_of_rooms': 2,'number_of_bathrooms':1, 'sqft': 1190}"
    }),
])

upload_predictor_params = OrderedDict([
    ('file', {
        'description': 'file',
        'type': 'file',
        'in': 'FormData',
        'required': True
    })
])

put_predictor_params = OrderedDict([
    ('name', {
        'description': 'The predictor name',
        'type': 'string',
        'in': 'path',
        'required': True
    }),
    ('data_source_name', {
        'description': 'The predictor name',
        'type': 'string',
        'in': 'body',
        'required': True
    }),
    ('from_data', {
        'description': '',
        'type': 'string',
        'in': 'body',
        'required': False
    }),
    ('to_predict', {
        'description': 'list of column names to predict',
        'type': 'array',
        'in': 'body',
        'required': True,
        'example': "['number_of_rooms', 'price']"
    })
])

# examples

EXAMPLES = [{
    'name': 'Price',
    'version': 1,
    'data_preparation': {
        'accepted_margin_of_error': 0.2,
        'total_row_count': 18000,
        'used_row_count': 10000,
        'test_row_count': 1000,
        'validation_row_count': 1000,
        'train_row_count': 8000
    },
    'data_analysis':  {
        'target_columns_metadata': [
            {
                'column_name': 'price',
                'data_type':  'numeric',
                'data_type_distribution': {
                    'type': 'categorical',
                    'x': ['numeric','categorical'],
                    'y': [19998,2]
                },
                'data_distribution': {
                    'data_histogram': {
                        'type': 'numeric',
                        'x': ['1000','1100','1200','1300','1400','1500','1600','1700','1800','1900','2000','2100', '2200', '2300', '2400'],
                        'y': [10,20,30,20,20,50,60,70,100,10,100,120,130,150,90]
                    },
                    'clusters': [
                        {
                            'group': 'low',
                            'members': ['1000','1100','1200','1300','1400','1500']
                        },
                        {
                            'group': 'medium',
                            'members': ['1600','1700','1800','1900']
                        },
                        {
                            'group': 'high',
                            'members': ['2000','2100', '2200', '2300', '2400']
                        },

                    ],
                    'mean': 1800
                },
                'consistency':  {
                    'score': '9/10',
                    'metrics': [
                        {
                            'type':  'warning',
                            'score': 0.3,
                            'description': 'There are two cells that dont match the numeric type, we will ignore them'
                        }
                    ],
                    'description': 'The data is in good shape, we have identified one consistency warning'
                },
                'redundancy': {
                    'score': '10/10'
                },
                'variability': {
                    'score': '8/10',
                    'metrics': [
                        {
                            'type':  'warning',
                            'score': 0.3,
                            'description': 'variability in this column was low, this means that there is a high bias towards properties in the 1k range'
                        },
                        {
                            'type':  'warning',
                            'score': 0.5,
                            'description': 'We have identified outliers in this column in the 30k price range'
                        }
                    ],
                    'description': 'The data is ok, we have identified one variability warnings'
                }

            }
        ],
        'input_columns_metadata': [
            {
                'column_name': 'sqft',
                'data_type':  'numeric',
                'data_type_distribution': {
                    'type': 'categorical',
                    'x': ['numeric'],
                    'y': [20000]
                },
                'data_distribution': {
                    'data_histogram': {
                        'type': 'numeric',
                        'x': ['100','150','200','250','300','350','400','450'],
                        'y': [10,20,30,20,20,50,60,70]
                    },
                    'mean': 1800
                },
                'consistency':  {
                    'score': '10/10'
                },
                'redundancy': {
                    'score': '10/10'
                },
                'variability': {
                    'score': '10/10'
                }

            },
            {
                'column_name': 'rooms',
                'data_type':  'numeric',
                'data_type_distribution': {
                    'type': 'categorical',
                    'x': ['numeric'],
                    'y': [20000]
                },
                'data_distribution': {
                    'data_histogram': {
                        'type': 'numeric',
                        'x': ['0','1','2','3','4','5'],
                        'y': [100,20,3000,200,2]
                    },
                    'mean': 2
                },
                'consistency':  {
                    'score': '10/10'
                },
                'redundancy': {
                    'score': '10/10'
                },
                'variability': {
                    'score': '9/10',
                    'metrics': [
                        {
                            'type':  'warning',
                            'score': 0.3,
                            'description': 'variability in this column was low, this means that there is a high bias towards properties in the 1k range'
                        },
                        {
                            'type':  'warning',
                            'score': 0.5,
                            'description': 'We have identified outliers in this column in the 30k price range'
                        }
                    ],
                    'description': 'The data is ok, we have identified one variability warnings'
                }

            }
        ]
    },
    'model_analysis': [
        {
            'column_name': 'price',
            'overall_input_importance': {
                'type': 'categorical',
                'x': ['sqft','rooms'],
                'y': [0.8, 0.2]
            },
            'train_accuracy_over_time':  {
                'type': 'numeric',
                'x': ['1','2', '3', '4', '5'],
                'y': [0.2, 0.2, 0.6, 0.8, 0.88]
            },
            'test_accuracy_over_time': {
                'type': 'numeric',
                'x': ['1','2', '3', '4', '5'],
                'y': [0.1, 0.12, 0.6, 0.8, 0.6]
            },
            'accuracy_histogram': {
                'x': ['1000','1800','1900', '2300', '2400'],
                'y': [0.9, 0.88, 0.99, 0.92, 0.34],
                'x_explained': [
                    [

                        {
                            'column_name': 'sqft',
                            'data_distribution': {
                                'data_histogram': {
                                    'type': 'numeric',
                                    'x': ['100','150','200','250','300','350','400','450'],
                                    'y': [10,0,0,0,2,5,6,7]
                                },
                                'mean': 1800
                            },
                            'consistency':  {
                                'score': '10/10'
                            },
                            'redundancy': {
                                'score': '10/10'
                            },
                            'variability': {
                                'score': '10/10'
                            }

                        },
                        {
                            'column_name': 'rooms',
                            'data_type':  'numeric',
                            'data_type_distribution': {
                                'type': 'categorical',
                                'x': ['numeric'],
                                'y': [20000]
                            },
                            'data_distribution': {
                                'data_histogram': {
                                    'type': 'numeric',
                                    'x': ['0','1','2','3','4','5'],
                                    'y': [0,0,300,20,100]
                                },
                                'mean': 2
                            },
                            'consistency':  {
                                'score': '10/10'
                            },
                            'redundancy': {
                                'score': '10/10'
                            },
                            'variability': {
                                'score': '9/10',
                                'metrics': [
                                    {
                                        'type':  'warning',
                                        'score': 0.3,
                                        'description': 'variability in this column was low, this means that there is a high bias towards properties in the 1k range'
                                    },
                                    {
                                        'type':  'warning',
                                        'score': 0.5,
                                        'description': 'We have identified outliers in this column in the 30k price range'
                                    }
                                ],
                                'description': 'The data is ok, we have identified one variability warnings'
                            }

                        }
                    ]

                ]


            }
        }
    ]
}]



import inspect
import os
from pathlib import Path
import traceback


def create_directory(path):
    path = Path(path)
    path.mkdir(mode=0o777, exist_ok=True, parents=True)

def get_paths():
    this_file_path = os.path.abspath(inspect.getfile(inspect.currentframe()))
    mindsdb_path = os.path.abspath(Path(this_file_path).parent.parent.parent)

    return [(f'{mindsdb_path}/etc/', f'{mindsdb_path}/var/predictors',f'{mindsdb_path}/var/datastore'),('/etc/mindsdb', '/var/lib/mindsdb/predictors','/var/lib/mindsdb/datastore'),('~/.local/etc/mindsdb','~/.local/var/lib/mindsdb/predictors','~/.local/var/lib/mindsdb/datastore')]

def get_or_create_dir_struct():
    for tup in get_paths():
        try:
            for dir in tup:
                assert(os.path.exists(dir))
                os.access(dir, os.W_OK)
            return tup[0], tup[1], tup[2]
        except Exception as e:
            pass

    for tup in get_paths():
        try:
            for dir in tup:
                create_directory(dir)
                os.access(dir, os.W_OK)
            return tup[0], tup[1], tup[2]
        except Exception as e:
            pass

    raise Exception(f'MindsDB storage directory: {path} does not exist and could not be created, trying another directory')



from mindsdb.api.mysql.mysql_proxy.datahub.information_schema import InformationSchema
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.mindsdb_datanode import MindsDBDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.mongo_datanode import MongoDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.csv_datanode import CSVDataNode


def init_datahub(config):
    all_ds = config['api']['mysql'].get('datasources', [])

    datahub = InformationSchema()

    datahub.add({
        'mindsdb': MindsDBDataNode(config)
    })

    csv_ds = [x for x in all_ds if x['type'].lower() == 'csv']
    for ds in csv_ds:
        datahub.add({
            ds['name']: CSVDataNode(ds['files'])
        })

    mongo_ds = [x for x in all_ds if x['type'].lower() == 'mongo']
    for ds in mongo_ds:
        datahub.add({
            ds['name']: MongoDataNode(ds['host'], ds['port'], ds['database'])
        })
    return datahub
